import argparse
import pathlib
import logging
import bencodepy

import Fiume.metainfo_decoder as md
import Fiume.state_machine as sm

##################################################

def main():
    options = sm.parser()
    
    with open(options["torrent_path"], "rb") as f:
        metainfo = md.MetaInfo(
            bencodepy.decode(f.read()) | options
        )

    tm = md.TrackerManager(metainfo, options)

    t = sm.ThreadedServer(
        metainfo, tm,
        **options
    )

    t.main()



import socket
import time
import threading
import logging
import enum
import os
import random
import pathlib 

from queue import Queue
from typing.io import *
from typing import *

import Fiume.config as config
import Fiume.utils as utils
import Fiume.master as master

logging.basicConfig(
    level=logging.DEBUG,
    format='[%(levelname)s] %(name)s %(asctime)s %(message)s',
    datefmt="%I:%M:%S"
)

class MexType(enum.Enum):
    HANDSHAKE = 84 # non cambiare
    KEEP_ALIVE = -1 #?
    CHOKE = 0
    UNCHOKE = 1
    INTERESTED = 2
    NOT_INTERESTED = 3
    HAVE = 4
    BITFIELD = 5
    REQUEST = 6
    PIECE = 7
    CANCEL = 8
    PORT = 9 # NOT USED

class Initiator(enum.Enum):
    SELF = 0
    OTHER = 1


#############################################


class PeerManager:
    def __init__(self, socket: Tuple,
                 metainfo, tracker_manager,
                 master_queues: Tuple[Queue, Queue],
                 initial_bitmap: List[bool],
                 options: Dict[str, Any],
                 initiator: Initiator):
        
        # Peer socket
        self.socket, self.address = socket
        self.peer_ip, self.peer_port = self.address
        # self.peer_ip, self.peer_port = self.socket.getsockname()
        # self.address = self.socket.getsockname() # TODO
        
        self.logger = logging.getLogger("TO " + str(self.peer_ip) + ":" + str(self.peer_port))
        self.logger.setLevel(options.get("debug_level", logging.DEBUG))
        self.logger.debug("__init__")
                
        self.metainfo = metainfo
        self.tracker_manager = tracker_manager

        self.options = options
        self.debug = False
        if "debug" in self.options:
            self.debug = self.options["debug"]
        
        # Ok
        self.peer_chocking, self.am_choking = True, True
        self.peer_interested, self.am_interested = False, False

        # Needed for establishing who starts the handshake
        self.initiator: Initiator = initiator
        self.received_handshake, self.sent_handshake = False, False

        # Queues for inter-thread communication
        self.queue_in, self.queue_to_master = master_queues

        # Bitmaps of my/other pieces
        self.my_bitmap = initial_bitmap
        self.peer_bitmap: List[bool] = utils.empty_bitmap(self.metainfo.num_pieces)

        # Output file
        self.out_fpath: pathlib.Path = self.initialize_file(self.options["output_file"])

        # Blocks that I don't have but my peer has
        self.am_interested_in: List[int] = list()
        self.peer_interested_in: List[int] = list()

        self.scheduled: List[int] = list()
        
        self.my_progresses: Dict[int, Tuple[bytes, int]] = dict()
        self.peer_progresses: Dict[int, Tuple[int, int]] = dict()

        self.cache_pieces: Dict[int, bytes] = dict()
        self.deferred_peer_requests: Dict[int, Tuple[int, int]] = dict()
        
        self.max_concurrent_pieces = 4
        
        self.old_messages: List[Tuple[str, bytes]] = list()
        self.completed = False

        
    def main(self):
        self.logger.debug("main")

        t1 = threading.Thread(target=self.message_socket_receiver)
        t1.start()
        
        # Stabilisce chi fra i due peers dovrà inviare il primo messaggio
        if self.initiator == Initiator.SELF:
            self.send_handshake()

        self.message_interpreter()

        
    def send_to_master(self, mex: utils.MasterMex):
        """
        Sends a message to the master's queue.
        """
        self.queue_to_master.put(mex)

        
    def read_data(self, piece_index, piece_offset=0, piece_length=0) -> bytes:
        """
        Reads data at a given offset from the cache. Used when peer
        asks me for a piece.

        If piece is not in cache, raise an exception (it should never happen! 
        every time you call read_data, data must be entered in cache!)
        """
        if piece_length == 0:
            piece_length = self.get_piece_size(piece_index)

        if piece_index not in self.cache_pieces:
            breakpoint()
            raise Exception(f"Don't have piece {piece_index} in cache!")

        piece = self.cache_pieces[piece_index]
        return piece[piece_offset:piece_offset+piece_length]

    
        
    def send_handshake(self):
        self.logger.debug("Sending HANDSHAKE")
        self.send_message(MexType.HANDSHAKE)
        self.sent_handshake = True
        if self.received_handshake:
            self.logger.debug("Sending BITFIELD")
            self.send_message(MexType.BITFIELD)


    def receive_handshake(self, mex):
        # Handshake ricevuta è corretta (info_hash matchano)
        assert mex[28:48] == self.metainfo.info_hash
        
        self.received_handshake = True
        self.logger.debug("Received HANDSHAKE")
        
        if self.sent_handshake:
            self.logger.debug("Sending BITFIELD")
            self.send_message(MexType.BITFIELD)
        else:
            self.send_handshake()        


    def initialize_file(self, fpath: pathlib.Path):
        """ Initialize the download file """
        if not fpath.exists():
            # TODO: BUG quando ad es. il file è /a/b/c/d.txt ma
            # le cartelle b e c non esistono
            fpath.touch()
            with open(fpath, "wb") as f:
                f.write(bytes(self.metainfo.num_pieces))
            return fpath

        return fpath

                                     
    # Thread a sé stante
    def message_socket_receiver(self):
        handshake_mex = self.socket.recv(68)

        self.queue_in.put(handshake_mex)
        
        while True:
            raw_length = self.socket.recv(4)
            length = int.from_bytes(raw_length, byteorder="big", signed=False)

            if length == 0:
                self.queue_in.put(b"")
                break
            
            raw_mex = bytes()
            while length != 0:
                data = self.socket.recv(length)
                raw_mex += data
                length -= len(data)
                if length < 0:
                    breakpoint()
                    _ = 0
                if length != 0 and self.debug:
                    print(f"Still waiting for {length} bytes...")

            self.queue_in.put(raw_length + raw_mex)

    
    def shutdown(self):
        self.logger.debug("Shutdown after receiving empty message from peer")
        self.send_to_master(utils.M_DISCONNECTED(self.address))
        # self.tracker_manager.notify_completion()
        exit(0)

    
    def try_unchoke_peer(self):
        if not self.am_choking:
            self.logger.debug("Asked if could unchoke peer, but it is already unchoked")
            return

        if not self.peer_interested:
            self.logger.debug("Asked if could unchoke peer, but peer is not interested")
            return

        # TODO: scrivere una funzione per valutare oggettivamente se ci sono
        # contrindicazioni nell'unchokare il peer. Per ora assumo si possa
        # sempre fare.
        self.am_choking = False
        self.send_message(MexType.UNCHOKE)
    
    #######

    def message_interpreter(self):
        """ Elabora un messaggio ricevuto, decidendo come rispondere e/o
        che cosa fare. """

        while True:
            mex = self.queue_in.get()

            # Messagi di controllo (ie. da Master) vengono intoltrati a questa funzione
            if isinstance(mex, utils.MasterMex):
                self.control_message_interpreter(mex)
                continue

            # TODO: messaggio vuoto b"" = fine scambio o errore di rete
            if mex == b"":
                print("Shutdown")
                self.shutdown()
                break
            
            try:
                mex_type = MexType(mex[4])
            except Exception as e:
                breakpoint()
                raise e

            self.logger.debug("Received message %s", str(mex_type))

            if mex_type == MexType.HANDSHAKE:
                self.receive_handshake(mex)

            elif mex_type == MexType.KEEP_ALIVE:
                self.send_message(MexType.KEEP_ALIVE)

            elif mex_type == MexType.CHOKE:
                self.peer_chocking = True
            elif mex_type == MexType.UNCHOKE:
                self.peer_chocking = False
                if self.am_interested:
                    # TODO: ricevuto unchoke, questa istruzione che segue
                    # richiede UN solo pezzo, invece che max_concurrent_pieces
                    self.try_ask_for_piece() 
            elif mex_type == MexType.INTERESTED:
                self.peer_interested = True
                self.try_unchoke_peer()
            elif mex_type == MexType.NOT_INTERESTED:
                self.peer_interested = False
                if not self.am_choking:
                    self.send_message(MexType.CHOKE)

            elif mex_type == MexType.HAVE:
                self.manage_received_have(utils.to_int(mex[5:9]))

            elif mex_type == MexType.BITFIELD:
                self.interpret_received_bitfield(mex[5:])

            elif mex_type == MexType.REQUEST:
                piece_index  = utils.to_int(mex[5:9]) 
                piece_offset = utils.to_int(mex[9:13]) 
                piece_length = utils.to_int(mex[13:17]) 
                self.manage_request(piece_index, piece_offset, piece_length)

            elif mex_type == MexType.PIECE:
                piece_index  = utils.to_int(mex[5:9]) 
                piece_offset = utils.to_int(mex[9:13])
                piece_payload = mex[13:]
                self.manage_received_piece(piece_index, piece_offset, piece_payload)

            elif mex_type == MexType.CANCEL:
                print("CANCEL not implemented")

            elif mex_type == MexType.PORT:
                print("PORT not implemented")

            else:
                print("ricevuto messaggio sconosciuto")
                breakpoint()

            
    def control_message_interpreter(self, mex: utils.MasterMex):
        self.logger.debug("[MASTER] Received message %s", str(type(mex)))
        
        if isinstance(mex, utils.M_KILL):
            self.logger.debug("[MASTER] Received KILL from master")
            self.send_to_master(utils.M_DEBUG("Got KILLED", self.address))
            return # TODO
        
        elif isinstance(mex, utils.M_DEBUG):
            self.logger.debug("[MASTER] Received DEBUG message from master: %s", mex.data)
            self.send_to_master(utils.M_DEBUG("Got DEBUGGED", self.address))
            return
        
        elif isinstance(mex, utils.M_OUR_BITMAP):
            self.logger.debug("[MASTER] Received OUR_BITMAP message from master")
            self.my_bitmap = mex.bitmap
            return
            
        elif isinstance(mex, utils.M_SCHEDULE):
            self.logger.debug("[MASTER] Received SCHEDULE message from master: %s", mex.pieces_index)
            self.scheduled += mex.pieces_index
            # Avoids deadlock:
            # if no pieces are currently downloaded, try_ask_for_piece will
            # start downloading the received piece(s);
            # if we were downloading pieces already, simply start one of the
            # new pieces (if we are not > max_concurrent_pieces
            self.ask_for_new_pieces()
            return

        
        elif isinstance(mex, utils.M_NEW_HAVE):
            self.logger.debug("[MASTER] Received NEW_HAVE message from master: %s", mex.piece_index)
            self.my_bitmap[mex.piece_index] = True
            return        

        # M_Piece has a bit complicated workflow.
        # If we receive a M_PIECE, it means that in the past we requested to the
        # master, on behalf of the peer, a piece N.
        # We put the entire piece N in a cache.
        # Then we resume the deferred request, 
        elif isinstance(mex, utils.M_PIECE):
            self.logger.debug("Received piece %d from master", mex.piece_index)
            
            self.cache_pieces[mex.piece_index] = mex.data

            if mex.piece_index not in self.deferred_peer_requests:
                # Should not happend
                self.logger.warning(
                    "Received piece %d from master, but we never deferred its sending!",
                    mex.piece_index
                )
                return

            # Resume request worfflow
            return self.manage_request(
                mex.piece_index,
                *self.deferred_peer_requests[mex.piece_index],
                deferred=True #important!
            )
                
        else:
            breakpoint()
            raise Exception("unknown message")
        
        
    def send_message(self, mexType: MexType, **kwargs):
        """
        Sends (via socket) a message to the peer.
        """
        if "delay" in self.options:
            time.sleep(self.options["delay"])
        self.socket.sendall(self.make_message(mexType, **kwargs))

        
    def make_message(self, mexType: MexType, **kwargs) -> bytes:
        """
        Builds a peer message of a given type (without sending it!)
        """
        mex = None
        
        if mexType == MexType.KEEP_ALIVE:
            mex = bytes([0,0,0,0])

        elif mexType == MexType.HANDSHAKE:
            mex = (utils.to_bytes(19) +
                    b"BitTorrent protocol" +
                    bytes(8) +
                    self.metainfo.info_hash +
                    utils.generate_peer_id(seed=self.peer_port))
        
        elif mexType.value in [0,1,2,3]:
            mex = (bytes([0,0,0,1]) +
                    utils.to_bytes(mexType.value, length=1))
        
        elif mexType == MexType.HAVE:
            mex = (utils.to_bytes(5, length=4) +
                    utils.to_bytes(mexType.value) +
                    utils.to_bytes(kwargs["piece_index"], length=4))
        
        elif mexType == MexType.BITFIELD:
            bitmap = utils.bool_to_bitmap(self.my_bitmap)

            mex = (utils.to_bytes(1 + len(bitmap), length=4) + 
                    utils.to_bytes(mexType.value) +
                    bitmap)

        elif mexType == MexType.REQUEST:
            mex = (utils.to_bytes(13, length=4) + 
                    utils.to_bytes(mexType.value) +
                    utils.to_bytes(kwargs["piece_index"], length=4) +
                    utils.to_bytes(kwargs["piece_offset"], length=4) +
                    utils.to_bytes(kwargs["piece_length"], length=4))

        elif mexType == MexType.PIECE:
            payload = self.read_data(
                kwargs["piece_index"],
                kwargs["piece_offset"],
                kwargs["piece_length"]
            )
            
            mex = (utils.to_bytes(9 + len(payload), length=4) + 
                    utils.to_bytes(mexType.value) +
                    utils.to_bytes(kwargs["piece_index"], length=4) +
                    utils.to_bytes(kwargs["piece_offset"], length=4) +
                    payload)

        if mex is None:
            raise Exception("Messaggio impossibile da costruire")

        return mex
                    

    def interpret_received_bitfield(self, mex_payload: bytes):
        """ 
        Analyzes and reacts to a received bitmap.
        """

        # Sets peer_bitmap according to the mex received
        self.peer_bitmap = utils.bitmap_to_bool(
            mex_payload,
            num_pieces=self.metainfo.num_pieces
        )

        if len(self.my_bitmap) < 80:
            # Pretty view of my/peer bitmaps
            # TODO: rendilo funzione a se stante
            print("my:   |", end="")
            for my in self.my_bitmap:
                print("x" if my else " ", end="")
            print("\npeer: |", end="")
            for peer in self.peer_bitmap:
                print("x" if peer else " ", end="")
            print()

        # Sanity check
        assert len(self.my_bitmap) == len(self.peer_bitmap)

        
        # Idenitifico (if any) i pieces del mio peer che io non ho
        for i, (m,p) in enumerate(zip(self.my_bitmap, self.peer_bitmap)):
            if not m and p:
                self.am_interested_in.append(i)

        self.logger.debug("[MASTER] Sending M_PEER_HAS to master")
        # Informs master of peer's bitmap
        self.send_to_master(
            utils.M_PEER_HAS(
                self.am_interested_in,
                self.address,
                self.max_concurrent_pieces+1, #TODO: migliorabile!
            )
        )
        
        # Se, dal confronto fra la mia e l'altrui bitmap, scopro
        # che non mi interessa nulla di ciò che ha il peer, informalo che
        # sei NOT_INTERESTED
        if len(self.am_interested_in) == 0 or len(self.scheduled) == 0:
            self.logger.debug("Nothing to be interested in")
            if self.am_interested:
                self.am_interested = False
                self.logger.debug("Sending NOT_INTERESTED")
                self.send_message(MexType.NOT_INTERESTED)
            return
        
        if self.am_interested: # Se ero già interessato in precedenza, non fare nulla
            return

        # Altrimenti dichiara il tuo interesse
        self.am_interested = True
        self.logger.debug("Sending INTERESTED message")
        self.send_message(MexType.INTERESTED)

        self.ask_for_new_pieces()

        return

    
    def get_piece_size(self, piece_index: int) -> int:
        """ 
        Returns the length of a piece. 
        
        This is needed, because the last pieces of a torrent probably
        has an irregular size. We want to catch corner cases, so...
        this is it.
        """
        
        if piece_index != self.metainfo.num_pieces - 1:
            return self.metainfo.piece_size

        last_piece_size = (
            self.metainfo.total_size - 
            (self.metainfo.num_pieces - 1) * self.metainfo.piece_size            
        )

        if last_piece_size != 0:
            return last_piece_size
        
        return self.metainfo.piece_size


    def ask_for_single_piece(self, piece_idx: int):
        """
        Low-level routine that sends the request for a piece to the peer.
        """
        self.logger.debug("Asking for new piece, number %d", piece_idx)
            
        piece_length = min(
            self.metainfo.block_size,
            self.get_piece_size(piece_idx)
        )
        
        try:    
            self.send_message(
                MexType.REQUEST,
                piece_index=piece_idx,
                piece_offset=0,
                piece_length=piece_length
            )
        except Exception as e:
            breakpoint()
            raise e

        # Inform the master that I have requested a new piece"
        self.send_to_master(
            utils.M_DEBUG("Requested new piece {piece_idx}", (self.peer_ip, self.peer_port))
        )
        
        # self.get_piece_size serve per gestire len irregolare dell'ultimo piece
        self.my_progresses[piece_idx] = (b"", self.get_piece_size(piece_idx))

        
        
    def ask_for_new_pieces(self):
        """ 
        Richiedo un pezzo completamente nuovo, cioè non già in self.progresses.
        """

        if len(self.am_interested_in) == 0 or len(self.scheduled) == 0:
            if len(self.scheduled) == 0:
                self.logger.debug("No pieces pending!")
            self.logger.debug("Nothing to be interested in")
            self.logger.debug("Sending NOT-INTERESTED message")
            if self.am_interested:
                self.am_interested = False
                self.send_message(MexType.NOT_INTERESTED)
            return

        if not self.am_interested:
            self.logger.warning("Want to ask piece, but I'm not interested? Sending INTERESTED", )
            self.am_interested = True
            self.send_message(MexType.INTERESTED)
            if self.peer_chocking:
                return # wait for unchoke

        if self.peer_chocking:
            self.logger.debug("Wanted to ask a new piece, but am choked")    
            return

        not_yet_started = set(self.am_interested_in) - set(self.my_progresses.keys())
        not_yet_started = not_yet_started & set(self.scheduled)
        
        # Se tutti i pieces sono già stati avviati o completati
        if len(not_yet_started) == 0:
            self.logger.debug("No NEW pieces are requestable; abort")
            return

        # Se sto già scaricando il numero max di pieces contemporaneamente
        if len(self.my_progresses) > self.max_concurrent_pieces:
            self.logger.debug("Already downloading at the fullest")
            return

        
        random_piece = random.sample(
            list(not_yet_started), #non si può fare random choice su set()
            k=min(self.max_concurrent_pieces - len(self.my_progresses),
                  len(not_yet_started))
        )

        for piece in random_piece:
            self.ask_for_single_piece(piece)

    
    def try_ask_for_piece(self, suggestion=None):
        """ Differisce da ask_for_new_piece: mentre l'altro chiede un pezzo
        mai scaricato prima, questo potrebbe anche riprendere il download
        di un pezzo già iniziato. """
        if self.peer_chocking:
            self.logger.debug("Wanted to request a piece, but am choked")
            return
        
        if len(self.my_progresses) == 0: # se non ci sono pezzi incompleti
            return self.ask_for_new_pieces()

        if len(self.my_progresses) > self.max_concurrent_pieces:
            self.logger.debug("Already topping max concurrent requests")
            return

        if not self.am_interested:
            self.logger.warning("Want to ask piece %d, but not interested; sending INTERESTED")
            self.am_interested = True
            self.send_message(MexType.INTERESTED)
            if self.peer_chocking:
                return
            
        if suggestion is not None:
            piece_idx = suggestion
        else:
            piece_idx = random.choice(list(self.my_progresses.keys()))

        (data_already_downloaded, total_len) = self.my_progresses[piece_idx]
        offset_start = len(data_already_downloaded)
        self.logger.debug("Will continue with piece %d from offset %d", piece_idx, offset_start)
        
        self.send_message(
            MexType.REQUEST, 
            piece_index=piece_idx,
            piece_offset=offset_start,
            piece_length=min(self.metainfo.block_size, total_len - offset_start)
        )
        

    def manage_received_have(self, piece_index: int):
        self.logger.debug("Acknowledging that peer has new piece %d", piece_index)
        self.peer_bitmap[piece_index] = True
        self.logger.debug(
            "[MASTER] Sending M_PEER_HAS %d to master, ask schedule %d pieces",
            piece_index, 1
        )
        
        self.send_to_master(
            utils.M_PEER_HAS(
                [piece_index],
                self.address,
                schedule_new_pieces=1
            )
        )

        
    def manage_received_piece(self, piece_index, piece_offset, piece_payload):
        if self.my_bitmap[piece_index]:
            self.logger.warning(
                "Received fragment of piece %d offset %d, but I have piece it already (len: %d)",
                piece_index, piece_offset, len(piece_payload)
            )
            return

        
        # Aggiorna my_progersses
        old_data, piece_size = self.my_progresses[piece_index]

        if piece_offset < len(old_data):
            self.logger.warning("Duplicate block, received offset %d but expecting %d",
                                piece_offset, len(old_data))
            return # TODO: ?
                                
        self.logger.debug("Received payload for piece %d offset %d length %d: %s...%s",
                          piece_index, piece_offset, len(piece_payload),
                          piece_payload[:4], piece_payload[-4:])
                                
        new_data = old_data + piece_payload
        
        if len(new_data) == piece_size:
            self.logger.debug("Completed download of piece %d", piece_index)
            
            if not self.verify_hash(piece_index, new_data):
                raise Exception("Hashes not matching") #TODO

            print("Completed: {:.1f}%".format(
                100 * (1 + sum(int(x) for x in self.my_bitmap)) / len(self.my_bitmap)
            ))
                  
            del self.my_progresses[piece_index]

            self.logger.debug("Sending HAVE for piece %d to peer", piece_index)
            self.send_message(MexType.HAVE, piece_index=piece_index)
            
            self.logger.debug("Setting my bitfield for piece %d as PRESENT", piece_index)
            self.update_my_bitmap(piece_index, True)
            self.am_interested_in.remove(piece_index)

            # M_PIECE richiede anche un nuovo pezzo al Master
            self.scheduled.remove(piece_index)
            self.logger.debug("[MASTER] Sending M_PIECE for %d", piece_index)
            self.send_to_master(utils.M_PIECE(piece_index, new_data, self.address))

            # Finito un pezzo, iniziane uno NUOVO
            self.ask_for_new_pieces()
            return
                              
        self.my_progresses[piece_index] = (new_data, piece_size)
        self.try_ask_for_piece(suggestion=piece_index)


        
    def manage_request(self, p_index, p_offset, p_length, deferred=False):
        """ Responds to a REQUEST message from the peer. """
        if self.am_choking:
            self.logger.warning("Received REQUEST but am choking.")
            return

        log_str = "Resuming deferred " if deferred else "Received "
        self.logger.debug(log_str + "REQUEST for piece %d offset %d length %d",
                          p_index, p_offset, p_length)

        if not self.peer_interested:
            self.logger.warning("Was asked for piece %d, but to me peer is not interested", p_index)
            breakpoint()
            return
        
        if self.peer_chocking:
            self.logger.debug("Was asked for piece %d, but peer is chocking me", p_index)
            # breakpoint()
            # return

        if not self.my_bitmap[p_index]:
            self.logger.warning("Was asked for piece %d, but I don't have it", p_index)
            breakpoint()
            return

        # 1. Controlla se il pezzo è in cache_pieces
        # 2. Se sì, fai come al solito ma leggendo dalla cache_pieces
        # 3. Se no, chiedi pezzo al master; archivia il messaggio di REQUEST
        # in self.deferred_requests, quindi quando dal master arriva un messaggio
        # M_PIECE, finalmente rispondi al Peer

        if not p_index in self.cache_pieces:
            self.debug("[MASTER] Deferred response to REQUEST piece %d", p_index)
            self.send_to_master(
                utils.M_PEER_REQUEST(p_index, self.address)
            )
            self.deferred_peer_requests[p_index] = (p_offset, p_length)
            return

        if deferred:
            offs, length = self.deferred_peer_requests[p_index]
            if offs != p_offset or length != p_length:
                self.logger.error("Resuming deferred piece %d, but couldn't find deferred request!",
                                  p_index)
                breakpoint()
                raise Exception("Deferentiationalitation error")

            del self.deferred_peer_requests[p_index]
            
                
        self.send_message(
            MexType.PIECE,
            piece_index=p_index,
            piece_offset=p_offset,
            piece_length=p_length
        )

        
        # TODO: revisione di queste due righe
        # TODO: rendile una funzione, da chiamare ad ogni invio di piece
        if p_index in self.peer_progresses:
            (old_partial, old_total) = self.peer_progresses[p_index]
        else:
            (old_partial, old_total) = (0, self.get_piece_size(p_index))
            
        if old_partial + p_length < self.metainfo.block_size:
            self.peer_progresses[p_index] = (old_partial + p_length, old_total)
        else:
            if p_index in self.peer_progresses:
                del self.peer_progresses[p_index]
            else:
                self.peer_progresses[p_index] = (old_partial + p_length, old_total)


    def update_my_bitmap(self, piece_index, val: bool):
        self.my_bitmap[piece_index] = val
        with open(utils.get_bitmap_file(self.out_fpath), "w") as f:
            f.write("".join(["1" if piece else "0" for piece in self.my_bitmap]))

        if all(self.my_bitmap):
            self.logger.debug("Download completed!")
            self.completed = True
            # TODO: esci 

        
    def verify_hash(self, piece_index: int, data: bytes):
        sha = utils.sha1(data)

        are_equal = sha == self.metainfo.pieces_hash[piece_index]

        if are_equal:
            self.logger.debug("Calculated hash for piece %d matches with metainfo", piece_index)
        else:
            self.logger.warning("Hashes for piece %d DO NOT MATCH!", piece_index)
            breakpoint()

        return are_equal


#################################ÀÀ
       
# Questo oggetto gestisce le connessioni entrambi.
# Ogni nuova connessione viene assegnata ad un oggetto TorrentPeer,
# il quale si occuperà di gestire lo scambio di messaggi
class ThreadedServer:
    def __init__(self, metainfo, tracker_manager, **options):
        self.host = "localhost"
        self.peer = None
        self.options = options

        self.logger = logging.getLogger("ThreadedServer")
        self.logger.debug("__init__")
        
        self.metainfo = metainfo
        self.tracker_manager = tracker_manager

        port = options["port"]
        self.logger.debug("Server is binding at %s", (self.host, port))
        
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind((self.host, port))

        self.port = self.sock.getsockname()[1]
        if port == 0:
            self.logger.debug("Self port: %d", self.port)

        self.peers = self.options.get("suggested_peers", [])
        peers = self.tracker_manager.notify_start()
        for ip, pport in peers: #filters myself
            if ip in ["localhost", utils.get_external_ip()]:
                if self.port == pport:
                    continue
            self.peers.append((ip, pport))
        
        # La bitmap iniziale, quando il programma viene avviato.
        # Viene letta da un file salvato in sessioni precedenti, oppure
        # creata ad hoc.
        # TODO: serve davvero, qui? Spostare in metainfo?
        self.global_bitmap: List[bool] = utils.data_to_bitmap(
            self.options["output_file"],
            num_pieces=self.metainfo.num_pieces
        )
        
        self.max_peer_connections = 1
        self.active_connections = set()
        
        self.mcu = master.MasterControlUnit(self.metainfo, self.global_bitmap, self.options)
        self.master_queue = self.mcu.get_master_queue()

        self.pms = list()

        
    def main(self):       
        socket_listen_t = threading.Thread(target=self.listen)
        socket_listen_t.start()        

        self.mcu.main()
        
        print("Available peers:", self.peers)
        if self.peers == []:
            self.logger.debug("No peers currently available")
            return
        
        i = 0
        while i < min(self.max_peer_connections, len(self.peers)):
            ip, port = random.choice(self.peers)
            self.logger.debug("Chosen peer: %s:%s", ip, port)
            
            if (ip, port) in self.active_connections:
                self.logger.warning("Chosen an already connected peer")
                time.sleep(1)
                continue
            
            if ip == self.host or port == self.port: #TODO: sbagliato, peer può usare mia stessa porta
                self.logger.debug("Attempting to connect to myself (%s): abort", (ip, port))
                time.sleep(2)
                continue

            try:
                queues = (Queue(), self.master_queue)
                peer_manager = self.connect_as_client(ip, port, queues)
                self.active_connections.add((ip, port))
                self.mcu.add_connection_to(peer_manager)
                self.pms.append(peer_manager)
                i += 1
                
            except Exception as e:
                self.logger.error("%s", e)
                time.sleep(2)
                continue
            
    
    def listen(self):
        self.logger.debug("Started listening on %s", (self.host, self.port))
        self.logger.debug("Max connections number: %d", self.max_peer_connections)
        
        self.sock.listen(self.max_peer_connections) # Numero massimo di connessioni in attesa (?)

        while True:
            self.logger.debug("Waiting for connections...")
            client_socket, address = self.sock.accept()
            self.logger.debug("Received connection request from: %s", address)
            
            newPeer = PeerManager(
                (client_socket, address),
                self.metainfo,
                self.tracker_manager,
                (Queue(), self.master_queue),
                self.global_bitmap,
                self.options,
                Initiator.OTHER
            )

            self.active_connections.add(address)
            self.mcu.add_connection_to(newPeer)
            self.pms.append(new_peer)
            
            t = threading.Thread(target = newPeer.main)
            t.start()

            
    def connect_as_client(self, ip, port, queues: Tuple[Queue, Queue]):
        new_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM) 
        new_socket.connect((ip, port))

        self.logger.debug("Actively connecting to %s", (ip, port))
        
        newPeer = PeerManager(
            (new_socket, (ip, port)),
            self.metainfo,
            self.tracker_manager,
            queues,
            self.global_bitmap,
            self.options,
            Initiator.SELF
        )

        t = threading.Thread(target = newPeer.main)
        t.start()            
        return newPeer

###############################

import Fiume.metainfo_decoder as md
import bencodepy
import argparse

def parser(s=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("torrent_path",
                        type=pathlib.Path,
                        help="path to .torrent file")

    parser.add_argument("output_file",
                        type=pathlib.Path,
                        help="where to download the file")

    parser.add_argument("-p", "--port",
                        action="store",
                        type=int,
                        default=50146,
                        help="port for this client")

    parser.add_argument("-v", "--verbosity",
                        action="count",
                        default=0,
                        dest="debug_level",
                        help="debug level")

    parser.add_argument("--suggested_peers",
                        action="store",
                        default=[],
                        help="suggested_peers")

    parser.add_argument("--delay",
                        type=float,
                        default=0,
                        help="delay for every sent message (only debug)")

    if s:
        options = vars(parser.parse_args(s))
    else:
        options = vars(parser.parse_args())
        
    options["debug_level"] = utils.int_to_loglevel(options["debug_level"])
    options["debug"] = False

    return options


if __name__ == "__main__":
    options = {
        "torrent_path": pathlib.Path("/home/groucho/torrent/image.jpg.torrent"),
        "output_file":  pathlib.Path("/home/groucho/torrent/downloads/image.jpg"),
        "delay": 0,
        "debug": False,
        "debug-level": logging.WARNING,
        "port": 50146,
        "suggested_peers": [("78.14.24.41", 50144)]
    }

    with open(options["torrent_path"], "rb") as f:
        metainfo = md.MetaInfo(
            bencodepy.decode(f.read()) | options
        )

    tm = md.TrackerManager(metainfo, options)

    t = ThreadedServer(
        metainfo, tm,
        **options
    )

    t.main()



import threading
import logging 

from queue import Queue
from typing import *

from Fiume.utils import *

logging.basicConfig(
    level=logging.DEBUG,
    format="[%(levelname)s] [%(name)s] %(message)s"
)

class ConnectionStatus:
    def __init__(self, peer):
        self.queue_in: Queue    = peer.queue_in
        self.peer_has: Set[int] = set()
        self.already_scheduled: Set[int] = set()

        
    def update_peer_has(self, pieces: List[int]):
        """
        Updates local list of pieces possessed by the peer.
        """
        self.peer_has |= set(pieces)

    def set_suggested(self, pieces: List[int]):
        self.already_scheduled |= set(pieces)
        
    def not_yet_scheduled(self) -> Set[int]:
        """
        Returns all the pieces that the peer has and that 
        were not already scheduled for request.
        """
        return self.peer_has - self.already_scheduled

    
    def completed_piece(self, piece_idx: int):
        """ 
        When we receive the full piece, remove it from the 
        already_scheduled set.
        """
        self.already_scheduled -= {piece_idx}


        
class MasterControlUnit:
    def __init__(self, metainfo, initial_bitmap, options):
        self.logger = logging.getLogger("Master")
        
        self.metainfo = metainfo
        self.options  = options
        self.bitmap: List[bool] = initial_bitmap
        
        self.connections: Dict[Address, ConnectionStatus] = dict()
        self.queue_in = Queue()

        # To prohbit concurrent access to the download file
        self.lock_download_file = threading.Lock()

        
    def get_master_queue(self) -> Queue():
        """
        Returns the queue with which communicate with master.
        """
        return self.queue_in
    
        
    def add_connection_to(self, peer: "PeerManager"):
        """
        Call this when you connect to a new peer.
        """
        self.connections[peer.address] = ConnectionStatus(peer)
        self.send_to(peer.address, M_OUR_BITMAP(self.bitmap))
        
        
    def send_to(self, address: Address, mex: MasterMex):
        """ 
        Sends a message to a peer manager, through the appropriate queue.
        """
        self.connections[address].queue_in.put(mex)

        
    def send_all(self, mex, exclude: Address=None):
        """
        Sends a message to every connected peer.
        """
        for p in self.connections:
            if p != exclude:
                self.send_to(p, mex)

                
    def update_global_bitmap(self, new_piece: int, peer_from: Address):
        """
        When receiving PIECE message, updates the global bitmap.
        Must also inform all peers of this update!
        """
        self.bitmap[new_piece] = True

        update_bitmap_file( # in utils.py
            # self.options["output_file"],
            self.metainfo.download_fpath,
            self.bitmap
        )

        self.send_all(M_NEW_HAVE(new_piece), peer_from)
        

    def bitmap_to_set(self) -> Set[int]:
        # TODO: sposta in utils
        out = set()
        for i in range(len(self.bitmap)):
            if self.bitmap[i]:
                out.add(i)
        return out

    
    def already_scheduled(self) -> Set[int]:
        """
        Returns all the pieces already scheduled by any peerManager.
        """
        return set.union(
            *[state.already_scheduled for state in self.connections.values()]
        )

    
    def schedule_for(self, address: Address, n=10) -> List[int]:
        """ 
        Schedules pieces to requests for a peer, taking into accounts
        the scheduled pieces for all other peers. 
        """
        state = self.connections[address]

        # Candidates pieces for peer P are those that:
        # 1. P owns
        # 2. Were not already assigned to PeerManager for P
        # 3. Were not already assigned to /any/ peerManager
        candidates_pieces = (
            state.not_yet_scheduled() -
            self.already_scheduled() -
            self.bitmap_to_set()
        )

        if len(candidates_pieces) == 0:
            # print("No candidates found...")
            return []
        
        chosen = random.sample(
            list(candidates_pieces),
            k = min(n, len(candidates_pieces))
        )

        state.set_suggested(chosen)
        
        return chosen


    def redistribute_pieces_of(self, address: Address) -> Dict[Address, List[int]]:
        """
        The goal is to redistribute the already scheduled pieces of
        peer P to all the other peers.
        
        We build a table of possible peers to which redistribute 
        the pieces; if more than one choice is possible, simply choose
        randomly.
        """
        state = self.connections[address]
        redistrib_pieces = state.already_scheduled
        
        # (piece_index da redistribuire) : (possibili peers che ce l'hanno) 
        table: Dict[int, List[Address]] = {
            p : [] for p in redistrib_pieces
        }
        
        for peer, p_state in self.connections.items():
            if peer == address:
                continue
            
            for common_piece in (p_state.peer_has & redistrib_pieces):
                table[common_piece].append(peer)

                
        new_assignments: Dict[Address, List[int]] = dict()
        
        for piece, candidate_peers in table.items():
            if candidate_peers == []:
                continue
            
            if len(candidate_peers) == 1:
                candidate = candidate_peers[0]
            else:
                candidate = random.choice(candidate_peers)

            new_assignments[candidate] = (
                new_assignments.get(candidate, []) + [piece]
            )

        return new_assignments

    
    def write_piece_to_file(self, piece_index: int, data: bytes):
        """
        Writes an entire piece, received from a peer, to the downloaded
        file. 

        It assumes that the data received where already hash-verified by
        the peer manager! 
        """
        with self.lock_download_file:

            if not self.metainfo.download_fpath.exists():
                self.metainfo.download_fpath.touch()

            with open(self.metainfo.download_fpath, "r+b") as f:
                # TODO: assert su lungehzza data
                f.seek(piece_index * self.metainfo.piece_size)
                f.write(data)

                
    def read_piece_from_file(self, piece_index) -> bytes:
        """
        Reads an entire piece from the downloaded file.
        """
        with self.lock_download_file:

            with open(self.metainfo.download_fpath, "r+b") as f:
                f.seek(piece_index * self.metainfo.piece_size)
                return f.read(self.metainfo.piece_size)


    
    
    def receiver_loop(self):
        while True:
            mex = self.queue_in.get()

            assert isinstance(mex, MasterMex)

            # When we are informed that a peer 
            if isinstance(mex, M_PEER_HAS):
                status = self.connections[mex.sender]
                status.update_peer_has(mex.pieces_index)

                answer = M_SCHEDULE(
                    self.schedule_for(mex.sender, n=mex.schedule_new_pieces)
                )
                self.send_to(mex.sender, answer)

                
            elif isinstance(mex, M_PIECE):
                status = self.connections[mex.sender]
                status.completed_piece(mex.piece_index)

                self.write_piece_to_file(mex.piece_index, mex.data)
                self.update_global_bitmap(mex.piece_index, mex.sender)                
                self.send_to(mex.sender,
                             M_SCHEDULE(self.schedule_for(mex.sender, n=mex.schedule_new_pieces)))

                # When completed (= we have all the pieces), inform all peers that we
                # have completed the download. The peers will decide if mantaining the
                # connection and seed, or to disconnect
                if all(self.bitmap):
                    self.send_all(M_DEBUG("completed", None))
                    print("Completed download!")

                    
            elif isinstance(mex, M_DISCONNECTED):
                mapping = self.redistribute_pieces_of(mex.sender)
                self.send_to(mex.sender, M_KILL())
                del self.connections[mex.sender]

                for peer_addr, new_scheduled in mapping.items():
                    self.connections[peer_addr].set_suggested(new_scheduled)                    
                    self.send_to(peer_addr, M_SCHEDULE(new_scheduled))

                    
            elif isinstance(mex, M_PEER_REQUEST):
                # TODO: un'idea. Al posto che inviare il pezzo intero sulla queue
                # (che è pesante, visto che un pezzo può essere anche 1Mb)
                # invia al peer il Lock usato per scrivere/leggere il file,
                # self.lock_download_file! 
                if not self.bitmap[mex.piece_index]:
                    self.send_to(mex.sender, M_ERROR(mex, "We don't have requested piece"))
                    continue
                
                data = self.read_piece_from_file(mex.piece_index)
                self.send_to(mex.sender,
                             M_PIECE(mex.piece_index, data, None, None))

                
            elif isinstance(mex, M_KILL):
                # Only debug, or user input
                break


    def main(self):
        t = threading.Thread(target=self.receiver_loop)
        t.start()
        



import bencodepy
import hashlib
import ipaddress
import random
import requests
# import multiprocessing as mp
import pathos.multiprocessing as mp

from requests.exceptions import Timeout
from math import log2
from typing import *
    
import Fiume.config as config
import Fiume.utils as utils

import logging
logging.getLogger("urllib3").setLevel(logging.WARNING)

Url = str
Address = Tuple[str, int] # (ip, port)

#################################

class MetaInfo(dict):
    """ 
    NB: solo per Single File Mode. 
    
    Classe che contiene le informazioni del file .torrent
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pieces_hash = utils.split_in_chunks(self[b"info"][b"pieces"], 20)

        sha = hashlib.sha1()
        sha.update(bencodepy.encode(self[b"info"]))
        self.info_hash = sha.digest()

        self.trackers: List[Url]  = self.__gather_trackers()

        self.piece_size: int = self[b"info"][b"piece length"]
        self.block_size: int = 16384 #16kb, standard
        self.total_size: int = self[b"info"][b"length"]
        self.num_pieces: int = len(self.pieces_hash)

        self.download_fpath: Path = self["output_file"]
                
    def __gather_trackers(self) -> List[Url]:
        """ Unites all possible trackers in a single list, useless """
        trackers = [self[b"announce"]]
        
        if b"announce-list" in self:
            trackers += [l[0] for l in self[b"announce-list"]]

        return trackers

    
    
###################


class TrackerManager:
    def __init__(self, metainfo: MetaInfo, options: Dict[str, Any]):
        self.logger = logging.getLogger("TrackManager")
        self.logger.setLevel(options.get("debug_level", logging.DEBUG))
        self.logger.debug("__init__")

        self.options = options
        self.bitmap_file = utils.get_bitmap_file(self.options["output_file"])
        
        self.metainfo: MetaInfo = metainfo
        self.working_trackers: List[str] = list()

        self.peer_id = config.CLIENT_INFO + random.randbytes(12)
        self.tracker_ids: Dict[Url, bytes] = dict()
        
        self.peers: List[Address] = list()
        
        #self.notify_start()

        
    def tell_all_trackers(self, params) -> List[Optional[Tuple[Url, requests.Response]]]:
        """ 
        Tells something to all trackers in the .torrent file.
        """
        def __tell_tracker_curry(url):
            return self.__tell_tracker(url, params)
        
        pool    = mp.Pool(16 * mp.cpu_count())
        results = pool.map(
            lambda url: self.__tell_tracker(url, params),
            self.metainfo.trackers
        )
        pool.close()
        return [r for r in results if r is not None]

    
    def __tell_tracker(self, url, params) -> Optional[Tuple[Url, requests.Response]]:
        """ 
        Single iterator for tell_all_trackers.
        """
        try:
            return (url, requests.get(url, params=self.base_params() | params, timeout=2.0))
        except Timeout:
            self.logger.debug("%s has time-outed", url)
            return None
        except Exception as e:
            self.logger.debug("%s has failed for some generic reason: %s", url, e) 
            return None

        
    def base_params(self) -> Dict:
        """ 
        Tracker GET request parameters that are always the same. Calculates `downloaded`,
        `uploaded` and `left` by reading the BITMAP file.
        """

        if self.bitmap_file.exists():
            with open(self.bitmap_file, "r") as f:
                bitmap = [int(c) for c in f.read().strip()]
                downloaded = sum(bitmap[:-1]) * self.metainfo.piece_size
                if bitmap[-1]:
                    downloaded += self.metainfo.total_size % self.metainfo.piece_size
                uploaded   = 0 # TODO
                left       = self.metainfo.total_size - downloaded
        else: # First connection 
            downloaded = 0
            uploaded = 0
            left = self.metainfo.total_size

            
        return {
            "info_hash": self.metainfo.info_hash,
            "peer_id": self.peer_id,
            "port": 50146,
            "compact": "1",
            "ip": "78.14.24.41",
            "downloaded": str(downloaded),
            "uploaded": str(uploaded),
            "left": str(left),
        }

    
    def notify_start(self, exclude_self=True) -> List[Address]:
        """ 
        Inform all the trackers that you are about to start downloading, and
        hence ask for peers.
        """
        self.logger.debug("Informing trackers I'm starting to download, asking for peers")
        self.logger.debug("%s", self.base_params())
        
        results = self.tell_all_trackers(
            {"event": "started"}
        )
        
        peers: Set[Address] = set()
        
        for tracker_url, __response in results:
            response = __response.content
            
            if response[:2] != b"d8":
                self.logger.debug("%s has returned a non-bencode object", tracker_url)
                continue
            
            self.working_trackers.append(tracker_url)

            response_bencode = bencodepy.decode(response)
            
            self.tracker_ids[tracker_url] = (
                response_bencode[b"tracker id"] if b"tracker id" in response_bencode else b""
            )

            peers = peers | self.__decode_peer(response_bencode)

        self.peers = list(peers)

        return list(peers)

    
    def notify_completion(self):
        """ 
        Inform all the trackers that you have finished downloading.
        In theory, you should call this /only/ when reaching 100%.
        """
        self.logger.debug("Notifying trackers of completion...")
        self.tell_all_trackers(
            {"event": "completed"}
        )

    def notify_stop(self):
        """
        Inform all the trackers that you are shutting down gracefully.
        """
        self.logger.debug("Notifying trackers of graceful shutdown...")
        self.tell_all_trackers(
            {"event": "stopped"}
        )

 
    def __decode_peer(self, response_bencode: bencodepy.Bencode) -> Set[Address]:
        """ 
        From a bencode bytestring to the list of peers' (ip, port).
        """
        
        peers = set()
        
        if not b"peers" in response_bencode:
            self.logger.debug("No peers in bencode answer from tracker")
            return set()
            
        for raw_address in utils.split_in_chunks(response_bencode[b"peers"], 6):
            ip = ipaddress.IPv4Address(raw_address[:4]).exploded
            port = utils.to_int(raw_address[4:6])
            peers.add((ip, port))

        self.logger.debug("Found the following peers: %s", str(peers))

        return peers

    
    def return_a_peer(self, exclude:List[int]=[]) -> Tuple[str, int]:
        return random.choice([p for p in self.peers if p[1] not in exclude])

    
def temp(b):
    peers = set()
    for raw_address in utils.split_in_chunks(b, 6):
        ip = ipaddress.IPv4Address(raw_address[:4]).exploded
        port = utils.to_int(raw_address[4:6])
        peers.add((ip, port))
    return peers



import pathlib
from Fiume.config import CLIENT_VERSION 
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="Fiume",
    version=".".join(CLIENT_VERSION.decode()),
    description="A Bittorrent client for single-file torrents.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/mattyonweb/fiume",
    author="Matteo Cavada",
    author_email="cvd00@insicuri.net",
    license="GNU General Public License v3.0",
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Topic :: Internet",
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3.9",
    ],
    packages=find_packages(),
    # package_dir={"": "Fiume"},
    install_requires=[
        "pathos", "requests", "bencode.py"
    ],
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "fiume=Fiume.cli:main",
        ]
    },
    python_requires='>=3.9',
)



import time
import queue
import threading
import enum

class Mex(enum.Enum):
    KILL = 0
    OK = 1

def mk_mex(payload):
    if payload is None:
        return Mex.KILL
    return payload

class CustomThread(threading.Thread):
    def __init__(self, group=None, target=None, name=None,
                 args=(), queue_in=None, queue_out=None):
        
        threading.Thread.__init__(self, group=group, target=target, name=name)
        
        self.queue_in = queue_in
        self.queue_out = queue_out

        self.kill_event = threading.Event()
        
    def receiver(self):
        """ Receives message from queue_in """
        while True:
            mex = self.queue_in.get()            

            if mex == Mex.KILL:
                self.shutdown()
                return

            self.dispatch(mex)

    def shutdown(self):
        print("Shutdown!")
        self.kill_event.set()
        exit(0)
        
    def sender(self, mex):
        """ Puts messages in queue_out """
        self.queue_out.put(mex)
        
    def dispatch(self, mex):
        print("Received", mex)
        
    def run(self):
        consumer_thread = threading.Thread(target=self.receiver)
        consumer_thread.start()

        i = 0
        while True:
            print(f"Waiting: {i} seconds")
            time.sleep(1)
            i += 1
            if self.kill_event.is_set():
                break
            
        consumer_thread.join()
        return

q1, q2 = queue.Queue(), queue.Queue()

t1 = CustomThread(queue_in=q1, queue_out=q2)
    
t1.start()

def porcodue(q):
    time.sleep(0.5)
    
    for i in range(2):
        print(i)
        q.put(i)
        time.sleep(1)
    q.put(Mex.KILL)

    time.sleep(3)
    t1.shutdown()

t2 = threading.Thread(target=porcodue, args=(q1,))
t2.start()

t2.join()
t1.join()

    



from unittest.mock import Mock, MagicMock
from queue import Queue
from pathlib import * 
from hashlib import sha1

import unittest
import random
import tempfile

from Fiume.utils import *
import Fiume.master as master

def repeat(times):
    def repeatHelper(f):
        def callHelper(*args):
            for diocane in range(0, times):
                f(*args)

        return callHelper

    return repeatHelper


class PeerHasEverything(unittest.TestCase):

    def setUp(self):
        piece_size = 256
        piece_number = 100

        self.data = [random.randbytes(piece_size) for _ in range(piece_number)]
        self.hashes = [sha1(piece) for piece in self.data]
        
        self.file = tempfile.NamedTemporaryFile()
        self.file.close()

        self.metainfo = Mock(
            piece_size=piece_size,
            piece_number=piece_number,
            download_fpath=Path(self.file.name),
            pieces_hash = self.hashes
        )
        
        self.peer  = Mock(address=("localhost", 50154), queue_in=Queue())
        self.peer2 = Mock(address=("localhost", 50155), queue_in=Queue())
        self.peer3 = Mock(address=("localhost", 50157), queue_in=Queue())

        self.initial_bitmap = [False for _ in range(piece_number)]
        
        self.mcu = master.MasterControlUnit(
            self.metainfo,
            self.initial_bitmap,
            {"download_fpath": Path("/dev/null")}
        )
        self.mcu.main()
        self.mcu.add_connection_to(self.peer)
        
    def tearDown(self):
        self.mcu.queue_in.put(M_KILL())
        
    def send_mcu(self, message):
        """ Helper """
        self.mcu.queue_in.put(message)

    def get_mex(self, peer, timeout=1):
        return peer.queue_in.get(timeout)
    
    ##############################
    
    def test_when_i_have_nothing_and_peer_everything(self):
        """ 
        When newly connected to a peer, the master should assign the 
        PeerManager 10 pieces to download. 
        """
        self.get_mex(self.peer) # M_out_bitmap mex
        self.send_mcu(
            M_PEER_HAS(list(range(100)), self.peer.address, schedule_new_pieces=10)
        )

        mex_to_peer = self.peer.queue_in.get(timeout=1)

        # Schedula 10 pezzi
        self.assertEqual(len(mex_to_peer.pieces_index), 10)

        # Non schedula mai due volte lo stesso pezzo
        self.assertSequenceEqual(
            sorted(mex_to_peer.pieces_index),
            sorted(list(set(mex_to_peer.pieces_index)))
        )


    # @repeat(9) # setUp() -> test (9 times) -> tearDown()
    def test_dont_reask_already_scheduled_pieces(self):
        """
        When a whole piece is received, this piece is sent to the master;
        who will:
        1) inform all the peerManagers of the new piece, with a
        HAVE message, so that everyone can update its bitmap; and 
        2) update its own global bitmap
        3) assign a new piece to download to the peerManager
        """
        # Diocane, ignora queste righe
        m = self.get_mex(self.peer) # M_out_bitmap mex (maybe)

        self.send_mcu(
            M_PEER_HAS(list(range(100)), self.peer.address, schedule_new_pieces=10)
        )

        # Master assigns schedules these pieces for the PeerManager 
        scheduled_pieces = self.peer.queue_in.get().pieces_index

        # We pretend that one of them has been received
        random_piece = random.choice(scheduled_pieces) 
        self.send_mcu(
            M_PIECE(random_piece, self.data[random_piece], self.peer.address)
        )

        # We receive two messages, one is the HAVE and the next one is the
        # new scheduled piece to download
        new_have     = self.peer.queue_in.get(timeout=1)
        new_schedule = self.peer.queue_in.get(timeout=1)

        # Checks on HAVE
        self.assertIsInstance(new_have, M_NEW_HAVE)
        self.assertEqual(new_have.piece_index, random_piece)

        # Checks on SCHEDULE
        self.assertIsInstance(new_schedule, M_SCHEDULE)
        self.assertEqual(len(new_schedule.pieces_index), 1)
        self.assertNotIn(
            random_piece,
            new_schedule.pieces_index
        )
        self.assertNotIn(
            new_schedule.pieces_index,
            scheduled_pieces,
            new_schedule
        )
       

    def test_dont_ask_piece_already_scheduled_to_another_peer(self):
        """
        Peer1 has all pieces from [0..60];
        Peer2 has all pieces from [50..100];
        Master must not schedule, for peer2, any piece from 50..60
        """
        self.get_mex(self.peer) # M_out_bitmap mex

        self.mcu.add_connection_to(self.peer2)
        self.get_mex(self.peer2) # M_out_bitmap mex

        self.send_mcu(
            M_PEER_HAS(list(range(60)),
                       self.peer.address,
                       schedule_new_pieces=60)
        )
        self.send_mcu(M_PEER_HAS(list(range(50, 100)), self.peer2.address))

        p1_scheduled = self.peer.queue_in.get()
        p2_scheduled = self.peer2.queue_in.get()

        self.assertFalse(
            any(x in range(50, 60) for x in p2_scheduled.pieces_index),
            p2_scheduled.pieces_index
        )


        
    def test_when_no_blocks_to_assign_assign_nothing(self):
        self.get_mex(self.peer) # M_out_bitmap mex
        self.send_mcu(M_PEER_HAS([0], self.peer.address, schedule_new_pieces=3))

        scheduled = self.peer.queue_in.get().pieces_index[0]

        self.assertEqual(scheduled, 0)

        self.send_mcu(M_PIECE(0, self.data[0], self.peer.address))

        _ = self.peer.queue_in.get() # HAVE message

        self.assertEqual(
            self.peer.queue_in.get().pieces_index, []
        )

        
    def test_graceful_disconnection_of_peer(self):
        """
        When a peer gracefully disconnects, the scheduled pieces that it had
        are redistributed to all the other pieces.
        """
        self.get_mex(self.peer) # M_out_bitmap mex

        # A peer reserves for itself all the pieces
        self.send_mcu(M_PEER_HAS(list(range(100)), self.peer.address, schedule_new_pieces=100))

        # A new peer tries to connect and ask for pieces, but
        # peer1 has already scheduled all the pieces; as a result,
        # no piece is scheduled for peer2
        self.mcu.add_connection_to(self.peer2)
        self.get_mex(self.peer2) # M_out_bitmap mex
        self.send_mcu(M_PEER_HAS(list(range(100)), self.peer2.address))
        self.assertEqual(
            self.get_mex(self.peer2).pieces_index,
            []
        )

        self.mcu.add_connection_to(self.peer3)
        self.get_mex(self.peer3) # M_out_bitmap mex
        self.send_mcu(M_PEER_HAS(list(range(50)), self.peer3.address))
        self.assertEqual(
            self.get_mex(self.peer3).pieces_index,
            []
        )

        # A new piece arrives from peer1
        self.send_mcu(M_PIECE(99, self.data[99], self.peer.address))

        # But now, peer1 disconnects!
        self.send_mcu(M_DISCONNECTED(self.peer.address))

        # HAVE messages, who cares
        _, _ = self.get_mex(self.peer2), self.get_mex(self.peer3)

        # SCHEDULE messages
        mex_peer2 = self.get_mex(self.peer2)
        mex_peer3 = self.get_mex(self.peer3)

        self.assertIsInstance(mex_peer3, M_SCHEDULE)
        self.assertIsInstance(mex_peer3, M_SCHEDULE)

        # No piece is scheduled to both the peers
        self.assertEqual(
            set(mex_peer2.pieces_index) & set(mex_peer3.pieces_index),
            set()
        )

        # Peer 3, who has only pieces [0..50], is not scheduled a piece
        # greater than 50
        self.assertTrue(
            all(x <= 50 for x in mex_peer3.pieces_index)
        )

        # All together, the two SCHEDULEs cover the scheduled pieces for P1
        self.assertSetEqual(
            set(mex_peer2.pieces_index) | set(mex_peer3.pieces_index),
            set(range(99))
        )

        
    def test_block_request(self):
        """ 
        Tests peer's requests for blocks.
        """
        self.get_mex(self.peer) # M_out_bitmap mex

        self.send_mcu(M_PEER_HAS(list(range(100)), self.peer.address, schedule_new_pieces=10))       
        scheduled = self.get_mex(self.peer).pieces_index
        random_piece = random.choice(scheduled)

        # A peer2 connects; it has no piece
        self.mcu.add_connection_to(self.peer2)
        self.get_mex(self.peer2) # M_out_bitmap mex
        self.send_mcu(M_PEER_HAS([], self.peer2.address))
        self.assertEqual(
            self.get_mex(self.peer2).pieces_index,
            []
        )

        # Peer2 asks for a piece that the master doesn't have;
        # an ERROR will be returned
        impossible_request = M_PEER_REQUEST(random_piece, self.peer2.address)
        self.send_mcu(impossible_request)
        
        error_p2 = self.get_mex(self.peer2)
        self.assertIsInstance(error_p2, M_ERROR)
        self.assertEqual(error_p2.on_service, impossible_request)

        
        # A piece arrives from peer1
        self.send_mcu(M_PIECE(random_piece,
                              self.data[random_piece],
                              self.peer.address))
        
        self.send_mcu(
            M_PEER_REQUEST(random_piece, self.peer2.address)
        )

        have_p2 = self.get_mex(self.peer2)
        self.assertIsInstance(have_p2, M_NEW_HAVE)
        self.assertEqual(have_p2.piece_index, random_piece)

        
        piece_p2 = self.get_mex(self.peer2)
        self.assertIsInstance(piece_p2, M_PIECE)
        self.assertEqual(piece_p2.piece_index, random_piece)
        self.assertEqual(piece_p2.data, self.data[random_piece])

        
    def test_completed(self):
        """ 
        When download is completed, master sends kill to all peers.
        """
        self.get_mex(self.peer) # M_out_bitmap mex

        self.send_mcu(M_PEER_HAS(list(range(100)), self.peer.address, schedule_new_pieces=100))       
        self.get_mex(self.peer)

        for i in range(100):
            self.send_mcu(M_PIECE(i, self.data[i], self.peer.address, schedule_new_pieces=0))
            have     = self.get_mex(self.peer) # scarta messaggio di risposta
            schedule = self.get_mex(self.peer)
            
        completed = self.get_mex(self.peer, timeout=1)
        self.assertIsInstance(completed, M_DEBUG)
        self.assertEqual(completed.data, "completed")

        
    def test_completed_but_exists_peer_with_missing_pieces(self):
        """ 
        When we complete, other peers should still be able to
        request pieces.
        """
        self.get_mex(self.peer) # M_out_bitmap mex

        self.send_mcu(M_PEER_HAS(list(range(100)), self.peer.address, schedule_new_pieces=100))       
        self.get_mex(self.peer)

        self.mcu.add_connection_to(self.peer2)
        self.get_mex(self.peer2) # M_out_bitmap mex
        self.send_mcu(M_PEER_HAS(list(range(25)), self.peer2.address))
        self.assertEqual(
            self.get_mex(self.peer2).pieces_index,
            []
        )
        
        for i in range(100):
            self.send_mcu(
                M_PIECE(i, self.data[i],
                        self.peer.address, schedule_new_pieces=0)
            )
            have     = self.get_mex(self.peer) # scarta messaggio di risposta
            schedule = self.get_mex(self.peer)

        for _ in range(100):
            self.get_mex(self.peer2) # HAVE messages

        self.send_mcu(M_PEER_REQUEST(99, self.peer2.address))
        m = self.get_mex(self.peer2)
        self.assertIsInstance(m, M_DEBUG)
        self.assertEqual(m.data, "completed")
        
        m = self.get_mex(self.peer2)
        self.assertIsInstance(m, M_PIECE)
        self.assertEqual(m.data, self.data[99])
